/*
 * bushfire_algorithm.h
 *
 * Template based implementation of a brushfire algorithm (bushfire since we are
 * in Aus) for a 2-dimensional grid. Supports multiple wavefronts (determined
 * by a wavefront type specifier). The priority queue provides the facility to
 * get the next element within the wavefront. It is prioritised by wavefront
 * type and then distance (note: wavefront type (the wavefront type must have
 * a less than ordering).
 *
 *
 *  Created on: 13/08/2011
 *      Author: daver
 */

#ifndef BUSHFIRE_H_
#define BUSHFIRE_H_

#include <map>
#include <tr1/unordered_set>
#include <tr1/unordered_map>
#include <boost/assert.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/foreach.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include "grid2d.h"
#include "mapmetadata_tools.h"
#include "Exceptions.h"

template<typename WaveType, typename DistanceMetric>
class BushFireQueue
{
  // A set of points
  typedef std::tr1::unordered_set<IntPoint2D> t_points;

  // Multimap - keyed by distance and stores a set of points
  typedef std::map<DistanceMetric, t_points> t_d2p;

  // Maintain a map of multimaps.
  typedef std::map<WaveType, t_d2p> t_map;

  t_map map_;
public:

  inline void push(const WaveType& wavetype, const DistanceMetric& dmpriority,
                   const IntPoint2D& point)
  {
    t_d2p& d2p = map_[wavetype];
    t_points& points = d2p[dmpriority];
    points.insert(point);
    //points.insert(typename t_points::value_type(dmpriority, point));
  }

  inline void pop()
  {
    // Get the first map based on wave type
    typename t_map::iterator it1 = map_.begin();
    if (it1 == map_.end())
    {
      BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("Cannot pop an empty queue")));
    }

    // Get the first set of points based on distance
    BOOST_ASSERT(!it1->second.empty());
    typename t_d2p::iterator it2 = it1->second.begin();

    // Get the first point
    BOOST_ASSERT(!it2->second.empty());
    typename t_points::iterator it3 = it2->second.begin();

    // Now erase
    it2->second.erase(it3);
    if (it2->second.empty())
    {
      it1->second.erase(it2);
      if (it1->second.empty())
      {
        map_.erase(it1);
      }
    }
  }

  /*
  typedef boost::tuple<const WaveType&, const DistanceMetric&, const UIntPoint2D&> t_front;

  inline t_front front() const
  {
    typename t_prioritymaps::iterator it1 = map_.begin();
    if (it1 == map_.end())
      {
        BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("No front of an empty queue")));
      }
    BOOST_ASSERT(!it1->second.empty());

    typename t_points::iterator it2 = it1->second.begin();
    return(t_front(it1->first, it2->first, it2->second));
  }
  */

  inline void front(WaveType& wt, DistanceMetric& dm, IntPoint2D& point) const
  {
    // Get the first map based on wave type
    typename t_map::const_iterator it1 = map_.begin();
    if (it1 == map_.end())
    {
      BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("No front of an empty queue")));
    }

    // Get the first set of points based on distance
    BOOST_ASSERT(!it1->second.empty());
    typename t_d2p::const_iterator it2 = it1->second.begin();

    // Get the first point
    BOOST_ASSERT(!it2->second.empty());
    typename t_points::const_iterator it3 = it2->second.begin();

    wt = it1->first;
    dm = it2->first;
    point = *it3;
  }

  inline bool empty() const
  {
    return map_.empty();
  }

  inline void clear()
  {
    map_.clear();
  }
};

template<typename Grid, typename WaveType, typename DistanceMetric> class BushFire;

// Need to declare the bushfire_update functions so they can be made
// friends of the Bushfire class.
template<typename Grid, typename WaveType, typename DistanceMetric, typename It>
void bushfire_update(BushFire<Grid, WaveType, DistanceMetric>& bushfire,
                     Grid& grid, It points_begin, It points_end);

template<typename Grid, typename WaveType, typename DistanceMetric>
void bushfire_update(BushFire<Grid, WaveType, DistanceMetric>& bushfire, Grid& grid);

template<typename Grid, typename WaveType, typename DistanceMetric>
class BushFire
{

public:
  typedef BushFire<Grid, WaveType, DistanceMetric> t_bushfire;
  typedef BushFireQueue<WaveType, DistanceMetric> t_queue;
  typedef enum { left=0x07, right=0x04, top=0x02, bottom=0x01,
                 topleft=0x09, topright=0x06, bottomleft=0x08,
                 bottomright=0x05, normal=0x00 } t_adjacency_type;

  typedef boost::function<void (t_bushfire&, const IntPoint2D&)> t_initialiser;
  typedef boost::function<void (t_bushfire&, const WaveType&, const DistanceMetric&,
                                const IntPoint2D&)> t_action;

private:
  std::vector<IntPoint2D> adjacency_;
  t_queue queue_;
  Grid* gridptr_;

  t_initialiser initialiser_;
  std::tr1::unordered_map<WaveType, t_action> actions_;

  template<typename It>
  friend void bushfire_update(t_bushfire& bushfire,
                              Grid& grid, It points_begin, It points_end);
  friend void bushfire_update<>(t_bushfire& bushfire, Grid& grid);

  // dodgy - should only be called by friend.
  void setGridPtr(Grid* grid)
  {
    gridptr_ = grid;
  }

  template<typename It>
  void init(It points_begin, It points_end)
  {
    queue_.clear();
    for (It it = points_begin; it != points_end; ++it)
    {
      initialiser_(*this, *it);
    }
  }

  void update()
  {
    // Loop through the queue and call the appropriate wavetype function
    while (!queue_.empty())
    {
      WaveType wavetype;
      DistanceMetric dist;
      IntPoint2D point;
      queue_.front(wavetype, dist, point);
      queue_.pop();
//      std::cout << "Handling wavetype: " << wavetype << " distance: " << dist << " point: " << point << std::endl;
      actions_[wavetype](*this, wavetype, dist, point);
    }
	std::cout << "Done" << std::endl;
  }

public:

  inline unsigned int maxX() const { return gridptr_->shape()[0]-1; }
  inline unsigned int maxY() const { return gridptr_->shape()[1]-1; }
  inline unsigned int sizeX() const { return gridptr_->shape()[0]; }
  inline unsigned int sizeY() const { return gridptr_->shape()[1]; }



  // Functions that the initialiser and different actions can use to
  // access the grid and queue. Note: these functions will only provide
  // meaningful data when called from within a bushfire_update call.

  inline t_queue& queue()
  {
    BOOST_ASSERT(gridptr_ != NULL);
    return queue_;
  }

  inline Grid& grid()
  {
    BOOST_ASSERT(gridptr_ != NULL);
    return *gridptr_;
  }

  t_adjacency_type adjacencyType(const IntPoint2D& point) const
  {
    BOOST_ASSERT(gridptr_ != NULL);
    unsigned char val = 0x00;
    if (point.x() == 0) val = val | left;
    else if (point.x() == maxX()) val = val | right;
    if (point.y() == 0) val = val | top;
    else if (point.y() == maxY()) val = val | bottom;
    return (t_adjacency_type) val;
  }

  inline bool isBorderCell(const IntPoint2D& point) const
  {
    return (adjacencyType(point) != normal);
  }

  inline typename Grid::element& cell(const IntPoint2D& point)
  {
    return grid()[point.x()][point.y()];
  }

  void adjacencyList(const IntPoint2D& point, std::vector<IntPoint2D>& adjacents)
  {
    int x; int y;
    for (int dx = -1; dx <= 1; ++dx)
    {
      x = dx + point.x();
      if (x < 0 || x > (int) maxX()) continue;
      for (int dy = -1; dy <= 1; ++dy)
      {
        if (dx == 0 && dy == 0) continue;
        y = dy + point.y();
        if (y < 0 || y > (int) maxY()) continue;
        adjacents.push_back(IntPoint2D((unsigned int)x, (unsigned int)y));
      }
    }
  }
  // Functions to initialise a bushfire object
  BushFire(t_initialiser initialiser): initialiser_(initialiser) {}

  // Initialise with a function object
  template<typename C, typename Fn>
  BushFire(C fo, Fn fn)
  {
    initialiser_ = boost::bind(fn, fo, _1, _2);
  }

  void setAction(const WaveType& wavetype, t_action& action)
  {
    actions_[wavetype] = action;
  }
  template<typename C, typename Fn>
  void setAction(const WaveType& wavetype, C fo, Fn fn)
  {
    actions_[wavetype] = boost::bind(fn, fo, _1, _2, _3, _4);
  }


  // NOTE: bushfire_update(BushFire<Grid, WaveType, DistanceMetric>& bushfire,
  //                        Grid& grid, It points_begin, It points_end)
  // does not seem to be linking properly. When I declare it as a friend
  // I need to provide the Iterator type and it fails to generate the proper instance.
  // Maybe I'm doing something illegal but this is only being picked up
  // by the linker. To get around this use the following member function
  // based interfaces.

  template<typename It>
  void update(Grid& grid, It points_begin, It points_end)
  {
    this->setGridPtr(&grid);
    this->init(points_begin, points_end);
    this->update();
    this->setGridPtr(NULL);
  }

  void update(Grid& grid)
  {
    std::vector<IntPoint2D> points;
    all_points(grid, points);
    this->setGridPtr(&grid);
    this->init(points.begin(), points.end());
    this->update();
    this->setGridPtr(NULL);
  }
};

// Update based on knowing which points on the grid have changed.
template<typename Grid, typename WaveType, typename DistanceMetric, typename It>
void bushfire_update(BushFire<Grid, WaveType, DistanceMetric>& bushfire,
                     Grid& grid, It points_begin, It points_end)
{
  bushfire.setGridPtr(&grid);
  bushfire.init(points_begin, points_end);
  bushfire.update();
  bushfire.setGridPtr(NULL);
}

// Update not knowing what has changed so have to consider all points in the grid.
template<typename Grid, typename WaveType, typename DistanceMetric>
void bushfire_update(BushFire<Grid, WaveType, DistanceMetric>& bushfire, Grid& grid)
{
  std::vector<IntPoint2D> points;
  all_points(grid, points);

  bushfire.setGridPtr(&grid);
  bushfire.init(points.begin(), points.end());
  bushfire.update();
  bushfire.setGridPtr(NULL);
}

#endif /* BUSHFIRE_H_ */
