/*
 * occupancygrid2d.h
 *
 *  Created on: 08/08/2011
 *      Author: daver
 */

#ifndef MAPMETADATA_TOOLS_H_
#define MAPMETADATA_TOOLS_H_

#include <limits>
#include <vector>
#include <list>
#include <ostream>
#include <cstdlib>
#include <utility>
#include <algorithm>
#include <functional>
#include <tr1/functional>
#include <tr1/unordered_set>
#include <boost/unordered_map.hpp>
#include <boost/functional/hash.hpp>
#include <boost/foreach.hpp>
#include "nav_msgs/MapMetaData.h"
#include "Exceptions.h"

class IntPoint2D : public std::pair<int, int>
{
public:
  IntPoint2D() : std::pair<int, int>(0, 0)
  {
  }

  IntPoint2D(int x, int y) :
    std::pair<int, int>(x, y)
  {
  }
  IntPoint2D(const IntPoint2D& o):
    std::pair<int, int>(o.x(), o.y())
  {
  }

  inline int x() const
  {
    return this->first;
  }
  inline int y() const
  {
    return this->second;
  }

  bool isNegative() const { return (x() < 0 || y() < 0); }


  inline unsigned int sqdistanceTo(const IntPoint2D& other) const
  {
    int dx = abs((int)x()- (int)other.x());
    int dy = abs((int)y()- (int)other.y());
    return (dx*dx + dy*dy);
  }

  inline float distanceTo(const IntPoint2D& other) const
  {
    float dx = (float)abs((int)x()- (int)other.x());
    float dy = (float)abs((int)y()- (int)other.y());
    return (std::sqrt(dx*dx + dy*dy));
  }

  // Calculate the points along a line (without the end points). Using
  // Bresenham's algorithm from:
  //    http://rosettacode.org/wiki/Bitmap/Bresenham%27s_line_algorithm#C.2B.2B
  inline void pointsTo(const IntPoint2D& other, std::list<IntPoint2D>& points,
                       bool inclusive=false) const
  {
    int x0 = x();
    int y0 = y();
    int x1 = other.x();
    int y1 = other.y();

    int dx = abs(x1-x0), sx = x0<x1 ? 1 : -1;
    int dy = abs(y1-y0), sy = y0<y1 ? 1 : -1;
    int err = (dx>dy ? dx : -dy)/2, e2;

    IntPoint2D np(x0,y0);
    if (inclusive) points.push_back(np);
    for(;;){
      np = IntPoint2D(x0,y0);
      if (np != *this && np != other) points.push_back(np);
      if (x0==x1 && y0==y1) break;
      e2 = err;
      if (e2 >-dx) { err -= dy; x0 += sx; }
      if (e2 < dy) { err += dx; y0 += sy; }
    }
    if (inclusive) points.push_back(np);
  }

  void shift(const IntPoint2D& sp)
  {
    this->first += sp.x();
    this->second += sp.y();
  }
};

struct VoronoiPoint2D
{
    IntPoint2D point;
    float distance;

    VoronoiPoint2D(){}
    VoronoiPoint2D(const IntPoint2D& p, float d) :
        point(p), distance(d) { }
    VoronoiPoint2D(const VoronoiPoint2D& other):
    point(other.point), distance(other.distance) { }
};


// For debugging purposes
bool breakpoint_check(const IntPoint2D&a, const IntPoint2D& b);

namespace std
{
  namespace tr1
  {

    inline std::size_t hash_value(const IntPoint2D& p)
    {
      std::size_t seed = 0;
      boost::hash_combine(seed, p.x());
      boost::hash_combine(seed, p.y());
      return seed;
    }

    inline std::size_t hash_value(const VoronoiPoint2D& vp)
    {
      std::size_t seed = 0;
      boost::hash_combine(seed, vp.point);
      boost::hash_combine(seed, vp.distance);
      return seed;
    }

    template<>
    struct hash<IntPoint2D>
    {
      inline std::size_t operator()(const IntPoint2D& p) const
      {
        return hash_value(p);
      }
    };

    template<>
    struct hash<VoronoiPoint2D>
    {
      inline std::size_t operator()(const VoronoiPoint2D& vp) const
      {
        return hash_value(vp);
      }
    };
  };
};



typedef std::tr1::unordered_set<IntPoint2D> t_pointset;
typedef boost::unordered_map<IntPoint2D, double> t_vpointmap;

inline t_pointset keys(const t_vpointmap& pmap)
{
    t_pointset tmp;
    BOOST_FOREACH(const t_vpointmap::value_type& vp, pmap)
    {
        tmp.insert(vp.first);
    }
    return tmp;
}

// Given a centre point project a tangent line out from the tangent point. The line will be
// projected in opposite directions with a distance dist. Returns these two end points from
// the projected line
std::pair<IntPoint2D, IntPoint2D> get_tangent_projection_points(const IntPoint2D& centre,
                                                                const IntPoint2D& tangent, float dist);


// Get the mid-point between two points
IntPoint2D get_midpoint(const IntPoint2D& a, const IntPoint2D& b);

void adjacents8(const IntPoint2D& point, std::vector<IntPoint2D>& adjlist);
void adjacents4(const IntPoint2D& point, std::vector<IntPoint2D>& adjlist);


inline std::ostream& operator<<(std::ostream& s, const IntPoint2D& point)
{
        return s << "(" << point.first << ", " << point.second << ")";
}

class IntSize2D : public std::pair<unsigned int, unsigned int>
{
public:
  IntSize2D() : std::pair<unsigned int, unsigned int>(0, 0)
  {
  }

  IntSize2D(unsigned int x, unsigned int y) :
    std::pair<unsigned int, unsigned int>(x, y)
  {
  }
  IntSize2D(const IntSize2D& o):
    std::pair<unsigned int, unsigned int>(o.x(), o.y())
  {
  }

  unsigned int x() const
  {
    return this->first;
  }
  unsigned int y() const
  {
    return this->second;
  }
};

typedef std::tr1::unordered_set<IntPoint2D> t_pointset;

class MapTransform
{
public:
//  typedef std::tr1::unordered_set<IntPoint2D> t_pointset;
private:

  bool needs_resize_;
  bool needs_shift_;

  IntPoint2D new_origin_;
  IntSize2D shift_size_;

  unsigned int xsize_;
  unsigned int ysize_;
  int xshift_;
  int yshift_;

  t_pointset newPoints_;

  static void allPoints(unsigned int xsize, unsigned int ysize, t_pointset& points);
  static void allPoints(unsigned int xsize, unsigned int ysize, const IntPoint2D& addto, t_pointset& points);

public:
  MapTransform(const nav_msgs::MapMetaData& orig_mmd,
               const nav_msgs::MapMetaData& new_mmd);

  bool needsResize() const;
  bool needsShift() const;

  const IntPoint2D& newOrigin() const;
  const IntSize2D& shiftWindowSize() const;

  const t_pointset& newPoints() const { return newPoints_; }
};
/*

  MapTransform(unsigned int xsize, unsigned int ysize,
               int xshift, int yshift):
                 xsize_(xsize), ysize_(ysize),
                 xshift_(xshift), yshift_(yshift){}
};
*/

// Calculates if a transform is required between a and b
bool requires_transform(const nav_msgs::MapMetaData& a, const nav_msgs::MapMetaData& b);

//UIntPoint2D get_origin(const nav_msgs::MapMetaData& mmd);


// Calculates how to transform a map conforming to 'a' to a map conforming to 'b'
std::auto_ptr<MapTransform> calculate_transform(const nav_msgs::MapMetaData& a,
                                                const nav_msgs::MapMetaData& b);


inline bool operator==(const geometry_msgs::Point& a, const geometry_msgs::Point& b)
{
    return (a.x == b.x && a.y == b.y && a.z == b.z);
}

inline bool operator!=(const geometry_msgs::Point& a, const geometry_msgs::Point& b)
{
    return !(a==b);
}


inline bool operator==(const geometry_msgs::Quaternion& a, const geometry_msgs::Quaternion& b)
{
    return (a.x == b.x && a.y == b.y && a.z == b.z && a.w == b.w);
}

inline bool operator!=(const geometry_msgs::Quaternion& a, const geometry_msgs::Quaternion& b)
{
    return !(a==b);
}

inline bool operator==(const geometry_msgs::Pose& a, const geometry_msgs::Pose& b)
{
    return (a.position == b.position && a.orientation == b.orientation);
}

inline bool operator!=(const geometry_msgs::Pose& a, const geometry_msgs::Pose& b)
{
  return !(a==b);
}


class FloatPoint2D : public std::pair<float, float>
{
public:
  FloatPoint2D() :
    std::pair<float, float>(0.0,0.0)
  {
  }

  FloatPoint2D(const FloatPoint2D& o) :
    std::pair<float, float>(o.x(), o.y())
  {
  }

  FloatPoint2D(float x, float y) :
    std::pair<float, float>(x, y)
  {
  }
  float x() const
  {
    return this->first;
  }
  float y() const
  {
    return this->second;
  }
};

/* limits provides a convenient source of some numeric limits.
 * A limits instance is only accessible through the static instance member
 * function - guaranteeing that only one instance will be created.
 */
class limits
{
public:
  float c_invalid_float_distance;
  int c_invalid_int_distance;
  IntPoint2D c_invalid_point;

  private:
  limits()
  {
/*
    if (!std::numeric_limits<float>::has_infinity)
    {
      BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("Infinity is not defined for type float")));
    }

    if (!std::numeric_limits<int>::has_infinity)
    {
      BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("Infinity is not defined for type int")));
    }
    if (!std::numeric_limits<int>::has_infinity)
    {
      BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("Infinity is not defined for type unsigned int")));
    }

    c_infinity_float = std::numeric_limits<float>::infinity();
    c_infinity_int = std::numeric_limits<int>::infinity();
    c_infinity_unsigned_int = std::numeric_limits<unsigned int>::infinity();
    c_infinity_point2d = IntPoint2D(c_infinity_unsigned_int, c_infinity_unsigned_int);
*/
    c_invalid_point = IntPoint2D(-1,-1);
    c_invalid_int_distance = -1;
    c_invalid_float_distance = -1;
  }

  static std::auto_ptr<limits> instance_;
  public:

  static const limits& instance()
  {
    if (instance_.get() == NULL)
    {
      std::auto_ptr<limits> tmp (new limits());
      instance_ = tmp;
    }
    return *instance_;
  }
};



#endif /* OCCUPANCYGRID2D_H_ */
