/*
 * max_voronoi_calc.h
 *
 *  Created on: 11/10/2011
 *      Author: daver
 */

#ifndef MAX_VORONOI_CALC_H_
#define MAX_VORONOI_CALC_H_

#include "mapmetadata_tools.h"
#include "bushfire2.h"
#include "voronoi.h"
#include "math.h"

// Starting from a given voronoi point follow the voronoi lines outwards (up to a radius threshold)
// and return the point which the furthest distance from its corresponding obstacle while still
// having its basis points sharing the wall of the original point.
class MaxVoronoiCalc
{
  typedef BushFire2<float, IntPoint2D> t_bushfire2;
  t_bushfire2 propagate_;
  const t_voronoimap& grid_;
  const t_pointset& boundary_points_;
  float wradius_;
  IntPoint2D source_;
  t_pointset visited_;
  std::vector<IntPoint2D> adjacents_;
  float maxObstDist_;
  IntPoint2D maxObstPoint_;
public:
  MaxVoronoiCalc(const t_voronoimap& grid, const t_pointset& boundary_points) :
	  propagate_(true), grid_(grid), boundary_points_(boundary_points){ }

  void process(t_bushfire2& bf, const float& distance, const IntPoint2D& point)
  {
    const VoronoiCell& cell = get_cell(grid_, point);
    const float point2obstDist = (point.distanceTo(cell.obst_) + point.distanceTo(cell.obst2_))/2.0;

    adjacents_.clear();
    adjacents8(grid_, point, adjacents_);
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
      const VoronoiCell& ac = get_cell(grid_, ap);
      if (!ac.isVoronoi()) continue;
      if (ac.isCriticalCorridor()) continue;
      if (boundary_points_.count(ac.obst_) == 0 || boundary_points_.count(ac.obst2_) == 0) continue;

      float dist2ap = source_.distanceTo(ap);
      if (dist2ap > wradius_ || dist2ap < distance) continue;
      if (visited_.find(ap) != visited_.end()) continue;
      visited_.insert(ap);
      float ap2obstDist = ap.distanceTo(ac.obst_);

      // move towards increasing obstacle distance - is too sensitive to
      // noise so add a bit of tolerance.
      float tolerabledist = point2obstDist - point2obstDist*0.05;
      if (ap2obstDist < tolerabledist) continue;

      if (ap2obstDist > maxObstDist_)
      {
        maxObstDist_ = ap2obstDist;
        maxObstPoint_ = ap;
      }
      propagate_.push(dist2ap, ap);
    }
  }

  const IntPoint2D& calculate(const IntPoint2D& point, float wradius)
  {
    source_ = point;
    wradius_ = wradius;
    visited_.clear();
    maxObstDist_ = point.distanceTo(grid_[point.x()][point.y()].obst_);
    maxObstPoint_ = point;
    propagate_.push(0.0, point);
    propagate_.update(t_bushfire2::make_callback(this, &MaxVoronoiCalc::process));
    return maxObstPoint_;
  }
};


#endif /* MAX_VORONOI_CALC_H_ */
