/*
 * opencv_tools.h
 *
 *  Created on: 10/08/2011
 *      Author: daver
 */

#ifndef OPENCV_TOOLS_H_
#define OPENCV_TOOLS_H_


#include <string>
#include <tr1/unordered_map>
#include <opencv2/opencv.hpp>
#include <boost/multi_array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/foreach.hpp>
#include <boost/assert.hpp>
#include <ros/ros.h>
#include "Exceptions.h"
#include "grid2d.h"
#include "map.h"
#include "basic_map.h"
#include "voronoi.h"
#include "voronoi_region_calc.h"

namespace cvtools
{

  class OccProbGrid2RGB
  {
    cv::Vec3b scalar_;

  public:
    inline const cv::Vec3b& operator()(signed char occval)
    {
      if (occval == 0) scalar_ = cv::Vec3b(255,255,255); // free space is white
//      else if (occval < 0) scalar_ = cv::Vec3b(255,150,150); // unknown is light blue
      else if (occval < 0) scalar_ = cv::Vec3b(175,175,175); // unknown is grey
      else
      {
        unsigned char inv = (100 - (unsigned char) occval);
//        unsigned char inv = occval;
        scalar_ = cv::Vec3b(inv, inv, inv); // rest is shades of grey
      }
      return scalar_;
    }
  };

  class OccMap2RGB
  {
    cv::Vec3b scalar_;

  public:
    inline const cv::Vec3b& operator()(const OccMapCell& mapval)
    {
      // BOOST_ASSERT(!mapval.isUnknown());
//      if (mapval.isFree()) scalar_ = cv::Vec3b(100,100,100); // grey
      if (mapval.isFree()) scalar_ = cv::Vec3b(255,255,255); // white
//      else if (mapval.isOccupied()) scalar_ = cv::Vec3b(0,0,0); // black
      else if (mapval.isOccupied()) scalar_ = cv::Vec3b(100,100,100); // light grey
//      else scalar_ = cv::Vec3b(255,150,150); // light blue
      else scalar_ = cv::Vec3b(0,0,0); // white
      return scalar_;
    }
  };

  class BasicMap2RGB
  {
    cv::Vec3b scalar_;

  public:
    inline const cv::Vec3b& operator()(const BasicMapCell& mapval)
    {
      if (mapval.isFree()) scalar_ = cv::Vec3b(255,255,255); // white
      else scalar_ = cv::Vec3b(0,0,0); // black
      return scalar_;
    }
  };

  class VoronoiMap2RGB
  {
    cv::Vec3b scalar_;

  public:
    inline const cv::Vec3b& operator()(const VoronoiCell& val)
    {
      int maxdistsq = 20*20;
      const limits& lim = limits::instance();

      // NOTE: The sqrt of the distance to the nearest obstacle may not be
      // set if there are no obstacles in the map!
      // BOOST_ASSERT(val.sqdist_ != lim.c_invalid_int_distance);

//      if (val.isOccupied())  scalar_ = cv::Vec3b(0,0,0); // black
      if (val.isOccupied())  scalar_ = cv::Vec3b(100,100,100); // light grey
      else if (val.isCriticalCorridor())
      {
        scalar_ = cv::Vec3b(0,0,255); // Dark Red
      }
/* DEBUG: Only show criticalcorridor candidates and corridor points for debugging
      else if (val.isCriticalCorridorCandidate())
      {
        scalar_ = cv::Vec3b(100,100,255); // Light Red
      }
      */
      else if (val.isCorridor())
      {
        scalar_ = cv::Vec3b(255,0,255); // Green?
      }
      else if (val.isVoronoi())
      {
//        scalar_ = cv::Vec3b(255,100,100); // blue
        scalar_ = cv::Vec3b(0,0,0); // blue
      }
      else
      {
    	  /*
        int sqdist = val.sqdist_;
        if (sqdist > maxdistsq) sqdist = maxdistsq;
        int col = (int) (((float)sqdist/float(maxdistsq))*100.0);
        col += 150;
        scalar_ = cv::Vec3b(col,col,col);
        */
    	  scalar_ = cv::Vec3b(255,255,255); // white
      //else (val.isFree()) scalar_ = cv::Vec3b(255,255,255); // white
      }
      return scalar_;
    }
  };


  // Note: I'm assuming that I have everything in an [x][y] grid. So the first dimension
  // corresponds to the width and second dimension the height. Also assume that [0][0]
  // is the bottom left corner.
  // OpenCV assumes the opposite. at(i,j) refers to the i'th row (ie height) and
  // j'th column (ie width) and (0,0) is at the top left.

  template<typename C>
  void check_grid_and_cvmat(const C& in_ma, const cv::Mat& out_cvmat)
  {
    if (out_cvmat.dims != 2 || in_ma.num_dimensions() != 2)
    {
      BOOST_THROW_EXCEPTION(UnsupportedFeature() << MessageErrInfo("Can only deal with 2D grids"));
    }
    if (out_cvmat.type() != CV_8UC3)
    {
      BOOST_THROW_EXCEPTION(UnsupportedFeature() << MessageErrInfo("Only supports RGB"));
    }
    // shape[0] - is width and shape[1] is height
    if ((int) in_ma.shape()[1] != out_cvmat.rows || (int) in_ma.shape()[0] != out_cvmat.cols)
    {
      BOOST_THROW_EXCEPTION(boost::enable_error_info(std::out_of_range("mis-match shape of arrays")));
    }
  }

  //template<typename ValueType, typename Fn>
  //void copy(const boost::multi_array<ValueType, 2>& in_ma, cv::Mat& out_cvmat,Fn fn)
  template<typename C, typename Fn>
  void copy(const C& in_ma, cv::Mat& out_cvmat, Fn fn)
  {
    check_grid_and_cvmat(in_ma, out_cvmat);
    int ma_width = in_ma.shape()[0];
    int ma_height = in_ma.shape()[1];

    for (int x = 0; x < ma_width; ++x)
    {
      for (int y = 0; y < ma_height; ++y)
      {
        const cv::Vec3b& rgb = fn(in_ma[x][y]);
        out_cvmat.at<cv::Vec3b>((ma_height-1)-y,x) = rgb;
      }
    }
  }

  template<typename C, typename Fn>
  void copy(const C& in_ma, const std::vector<IntPoint2D>& in_points, cv::Mat& out_cvmat,Fn fn)
  {
    check_grid_and_cvmat(in_ma, out_cvmat);
    int ma_height = in_ma.shape()[1];

    BOOST_FOREACH( const IntPoint2D& point, in_points)
    {
      unsigned int x = point.x();
      unsigned int y = point.y();
      const cv::Vec3b& rgb = fn(in_ma[x][y]);
      out_cvmat.at<cv::Vec3b>((ma_height-1)-y,x) = rgb;
    }
  }

  template<typename C, typename It, typename Fn>
  void copy(const C& in_ma, It begin, It end, cv::Mat& out_cvmat,Fn fn)
  {
    check_grid_and_cvmat(in_ma, out_cvmat);
    int ma_height = in_ma.shape()[1];

    while (begin != end)
    {
      IntPoint2D point = *begin;
      unsigned int x = point.x();
      unsigned int y = point.y();
      const cv::Vec3b& rgb = fn(in_ma[x][y]);
      out_cvmat.at<cv::Vec3b>((ma_height-1)-y,x) = rgb;
      ++begin;
    }
  }

  template<typename It>
  void render_voronoimap(const t_voronoimap& grid, It begin, It end, cv::Mat& out_cvmat)
  {
    typedef std::pair<IntPoint2D, IntPoint2D> t_line;
    typedef std::vector<t_line> t_lines;
    t_lines lines;
    t_pointset midpoints;
    t_pointset critpoints;

	// Define some colours
    cv::Scalar red = cv::Scalar(0,0,255); // Red
    cv::Scalar green = cv::Scalar(0,255,0); // Green
    cv::Scalar grey = cv::Scalar(50,50,50); // dark grey
    cv::Scalar blue = cv::Scalar(255,0,0); // Blue?



    // First do the points - but generate the lines from corridor points
    check_grid_and_cvmat(grid, out_cvmat);
    int ma_height = grid.shape()[1];
    VoronoiMap2RGB fn;

    while (begin != end)
    {
      IntPoint2D point = *begin;
      unsigned int x = point.x();
      unsigned int y = point.y();
      const VoronoiCell& cell = grid[x][y];
      if (cell.isCriticalCorridor())
      {
        lines.push_back(t_line(point, cell.obst_));
        lines.push_back(t_line(point, cell.obst2_));
        critpoints.insert(point);
      }
//      if (cell.forceg_.isMidP_) midpoints.insert(point);

      const cv::Vec3b& rgb = fn(cell);
      out_cvmat.at<cv::Vec3b>((ma_height-1)-y,x) = rgb;
      ++begin;
    }

    // Now draw the voronoi corridor lines
    BOOST_FOREACH(const t_line& line, lines)
    {
      cv::Point a(line.first.x(), (ma_height-1) - line.first.y());
      cv::Point b(line.second.x(), (ma_height-1) - line.second.y());
//      cv::line(out_cvmat,a,b,red);
//      cv::line(out_cvmat,a,b,grey,2.0);
      cv::line(out_cvmat,a,b,blue,2.0);
    }

    // Draw critical points
    BOOST_FOREACH(const IntPoint2D& p, critpoints)
    {
    	cv::Point ap(p.x(), (ma_height-1) - p.y());
        //cv::circle(out_cvmat,  ap, 5, grey);
    }

    // Now draw the mid points
    BOOST_FOREACH(const IntPoint2D& p, midpoints)
    {
      cv::Point cvp(p.x(), (ma_height-1)-p.y());
      cv::circle(out_cvmat,  cvp, 0, green);
      cv::circle(out_cvmat,  cvp, 5, green);
    }
  }



  // CVManager's manages a bunch of RGB images that reflect different views
  // of the 2D map.
  class CVManager
  {
    typedef std::tr1::unordered_map<std::string, boost::shared_ptr<cv::Mat> >t_name2mat;
    t_name2mat name2mat_;

    // Use the id for the mouse callback
    typedef std::tr1::unordered_map<unsigned int, std::string> t_id2name;
    t_id2name id2name_;

    unsigned int width_; // number columns
    unsigned int height_; // number rows
    int xorigin_;
    int yorigin_;

    std::vector<IntPoint2D> allpoints_;
    std::auto_ptr<t_voronoimap> vmap_;
    std::auto_ptr<VoronoiRegionMap> regionmap_;

    CVManager();

    static std::auto_ptr<CVManager> instance_;
  public:
    ~CVManager();

    static CVManager& instance();

    // Some generic functions - maybe should be private
    void resize(unsigned int width, unsigned int height, int xorigin, int yorigin);
    cv::Mat& getImage(const std::string& name);
    void showImage(const std::string& name);

    void updateOccProbGrid(const t_occgrid& oc, const std::vector<IntPoint2D>& in_points);
    void updateOccProbGrid(const t_occgrid& oc);

    void updateOccMap(const t_occmap& oc, const std::vector<IntPoint2D>& in_points);
    void updateOccMap(const t_occmap& oc);

    void updateOccMap2(const t_occmap& oc, const std::vector<IntPoint2D>& in_points, const VoronoiRegionMap& rm);
    void updateOccMap2(const t_occmap& oc,const VoronoiRegionMap& rm);


    void updateBasicMap(const t_basicmap& oc, const std::vector<IntPoint2D>& in_points);
    void updateBasicMap(const t_basicmap& oc);

    template<typename It>
    void update(const t_voronoimap& oc, It begin, It end, const VoronoiRegionMap& rm)
    {
      regionmap_ = rm.clone();
      vmap_ = create_grid2d<t_voronoimap, t_voronoimap>(oc);

      //cvtools::VoronoiMap2RGB filter;
      cv::Mat& img = getImage("Voronoi Map");
      render_voronoimap(oc, begin, end, img);
      cv::imshow("Voronoi Map", img);
    }

    void update(const t_voronoimap& oc, const VoronoiRegionMap& rm);

    // The OpenCV mouse callback function
    static void OnMouse(int event, int x, int y, int flags, void *param);

    void DisplayCell(const t_voronoimap& vmap, const IntPoint2D& point);
    void DisplayVoronoiCell(const VoronoiCell& vc, const IntPoint2D& point);

    void DisplayRegions(const std::string& imgname);
  };
};




#endif /* OPENCV_TOOLS_H_ */
