/*
 * voronoi.h
 *
 *  Created on: 11/08/2011
 *      Author: daver
 */

#ifndef VORONOI_H_
#define VORONOI_H_

#include <limits>
#include <map>
#include <boost/shared_ptr.hpp>
#include <boost/assert.hpp>
// #include <geos.h>
#include "bushfire.h"
#include "basic_map.h"
#include "map.h"
#include "connected_points.h"

// A lot of this is adapted/copied from dynamicvoronoi
class VoronoiCell
{
public:
  typedef enum { occupied=0, free=1, voronoi=2, freeVoronoiSkel=3, corridor=4,
                 criticalCorridorCandidate=5, criticalCorridor=6 } t_occupancy;

public:
  int sqdist_; // makes calculation faster.
  t_occupancy occupancy_; // state of the cell
  IntPoint2D obst_; // The nearest obstacle
  IntPoint2D obst2_; // A voronoi point will have (at least) a second nearest obstacle.

  IntPoint2D nearCorr_; // The nearestCorridor within the epsilon neighbourhood
  int ncdistpx_; // distance of the nearest corridor (along the voronoi line).

  // The ratio that defines a point as a criticalCorridor.
  // The higher the value the better. Note: the only reason to store this ratio
  // is for debugging purposes.
  float ccRatio_;

  // The variance (error margin) of the sum of the distance to the two obstacles
  // versus the distance between the two obstacles.
  float corrVariance_;

  IntPoint2D nearVoronoi_; // The nearest voronoi point that shares the same obstacle

  VoronoiCell(t_occupancy occupancy = occupied);
  VoronoiCell(const VoronoiCell& o);
  void validate(int minobstdistsq) const;

  bool isOccupied() const { return (occupancy_ == occupied); }
  bool isFree() const { return (occupancy_ >= free); }
  bool isVoronoi() const { return (occupancy_ >= voronoi); }
  bool isVoronoiSkel() const { return (occupancy_ >= freeVoronoiSkel); }
  bool isCorridor() const { return (occupancy_ >= corridor); }
  bool isCriticalCorridorCandidate() const { return (occupancy_ >= criticalCorridorCandidate); }
  bool isCriticalCorridor() const { return (occupancy_ >= criticalCorridor); }

  inline bool operator==(const VoronoiCell& o) { return (occupancy_ == o.occupancy_); }
  inline bool operator!=(const VoronoiCell& o) { return (occupancy_ != o.occupancy_); }

  void operator=(const t_occupancy& o);
  void operator=(const VoronoiCell& o);

};

typedef boost::multi_array<VoronoiCell,2> t_voronoimap;

/*
 * Functions to resize/copy/shift a t_voronoimap grid.
 * Because the VoronoiCells make reference to points within the grid itself,
 * therefore to copy and shift a map properly means that you have to
 * modify values of each cell.
 */

void voronoimap_validate(const t_voronoimap& vmap, int minObstDistSq);

std::auto_ptr<t_voronoimap> voronoimap_create(const t_basicmap& input,
                                              t_pointset& boundarypoints);

std::auto_ptr<t_voronoimap> voronoimap_create(const t_occmap& input,
                                              t_pointset& boundarypoints);

void voronoimap_copy_to(const t_voronoimap& input, t_voronoimap& output, const IntPoint2D& new_origin);

// Conversion functor to convert the basic map to a voronoi map
struct BasicMap2VoronoiMap
{
  inline VoronoiCell::t_occupancy operator()(const BasicMapCell& bmc)
  {
    if (bmc.occupancy_ == BasicMapCell::occupied) { return VoronoiCell::occupied; }
    else return VoronoiCell::free;
  }
};

// Conversion functor to convert the occ map to a voronoi map
struct OccMap2VoronoiMap
{
  inline VoronoiCell::t_occupancy operator()(const OccMapCell& omc)
  {
    if (omc.isOccupied()){ return VoronoiCell::occupied; }
    else if (omc.isFree()){ return VoronoiCell::free; }
    BOOST_ASSERT(false);
  }
};


#endif /* VORONOI_H_ */
