/*
 * occupancygrid2d.cpp
 *
 *  Created on: 08/08/2011
 *      Author: daver
 */

#include <sstream>
#include <ros/ros.h>
#include "Exceptions.h"
#include "mapmetadata_tools.h"
#include "grid2d.h"

// Given a centre point project a tangent line out from the tangent point. The line will be
// projected in opposite directions with a distance dist. Returns these two end points from
// the projected line
std::pair<IntPoint2D, IntPoint2D> get_tangent_projection_points(const IntPoint2D& centre,
                                                                const IntPoint2D& tangent, float dist)
{
  double const PI = 4*std::atan(1);

  float ydiff = tangent.y() - centre.y();
  float xdiff = tangent.x() - centre.x();
  float theta1 = std::atan2(ydiff, xdiff);
  float theta2 = (PI/2)-theta1;

  float dx = std::cos(theta2)*dist;
  float dy = std::sin(theta2)*dist;

  IntPoint2D tpp1(tangent.x() + (int) dx, tangent.y() - (int) dy);
  IntPoint2D tpp2(tangent.x() - (int) dx, tangent.y() + (int) dy);

  return make_pair(tpp1, tpp2);
}

// Get the mid-point between two points
IntPoint2D get_midpoint(const IntPoint2D& a, const IntPoint2D& b)
{
  float min_x = static_cast<float>(std::min(a.x(), b.x()));
  float min_y = static_cast<float>(std::min(a.y(), b.y()));
  float max_x = static_cast<float>(std::max(a.x(), b.x()));
  float max_y = static_cast<float>(std::max(a.y(), b.y()));

  float new_x = min_x + (max_x - min_x)/2;
  float new_y = min_y + (max_y - min_y)/2;
  return IntPoint2D(static_cast<int>(new_x), static_cast<int>(new_y));
}

void adjacents8(const IntPoint2D& point, std::vector<IntPoint2D>& adjlist)
{
  int x; int y;
  for (int dx = -1; dx <= 1; ++dx)
  {
    x = dx + point.x();
    for (int dy = -1; dy <= 1; ++dy)
    {
      if (dx == 0 && dy == 0) continue;
      y = dy + point.y();
      adjlist.push_back(IntPoint2D(x,y));
    }
  }
}

void adjacents4(const IntPoint2D& point, std::vector<IntPoint2D>& adjlist)
{
  int x; int y;
  for (int dx = -1; dx <= 1; ++dx)
  {
    x = dx + point.x();
    for (int dy = -1; dy <= 1; ++dy)
    {
      if (dx == 0 && dy == 0) continue;
      if (dx != 0 && dy != 0) continue;
      y = dy + point.y();
      adjlist.push_back(IntPoint2D(x,y));
    }
  }
}


// For debugging purposes
bool breakpoint_check(const IntPoint2D&a, const IntPoint2D& b)
{
  if (a == b)
  {
    std::cout << "Breakpoint check triggered at: " << a << std::endl;
    return true;
  }
  return false;
}


MapTransform::MapTransform(const nav_msgs::MapMetaData& orig_mmd,
                           const nav_msgs::MapMetaData& new_mmd)
{
  needs_resize_ = (orig_mmd.width != new_mmd.width || orig_mmd.height != new_mmd.height);
  needs_shift_ = false;
  if (!needs_resize_)
  {
    shift_size_ = IntSize2D(0,0);
    return;
  }
  newPoints_.clear();
  allPoints(new_mmd.width, new_mmd.height, newPoints_);

  // Shift the whole of the original map

  shift_size_ = IntSize2D(orig_mmd.width, orig_mmd.height);

  if (orig_mmd.origin.orientation != new_mmd.origin.orientation)
  {
    BOOST_THROW_EXCEPTION(UnsupportedFeature()
        << MessageErrInfo("No support for transforming the orientation of maps"));
  }
  if (orig_mmd.resolution != new_mmd.resolution)
  {
    BOOST_THROW_EXCEPTION(UnsupportedFeature()
        << MessageErrInfo("No support for transforming the resolution of maps"));
  }

  if (orig_mmd.width > new_mmd.width || orig_mmd.height > new_mmd.height)
  {
	  std::stringstream ss;
	  ss << "No support for shrinking maps: trying to shrink map "
	     << orig_mmd.width << "x" << orig_mmd.height << " to "
	     << new_mmd.width << "x" << new_mmd.height;
	  ROS_FATAL("%s", ss.str().c_str());
	  BOOST_THROW_EXCEPTION(UnsupportedFeature() << MessageErrInfo(ss.str()));
  }

  int xshift = 0;
  int yshift = 0;
  FloatPoint2D a_origin(orig_mmd.origin.position.x, orig_mmd.origin.position.y);
  FloatPoint2D b_origin(new_mmd.origin.position.x, new_mmd.origin.position.y);

  // Calculate any necessary shift
  if (a_origin != b_origin)
  {
    needs_shift_ = true;
    xshift = (int)((a_origin.x() - b_origin.x())/orig_mmd.resolution);
    yshift = (int)((a_origin.y() - b_origin.y())/orig_mmd.resolution);
  }
  if (xshift < 0 || yshift < 0)
  {
    BOOST_THROW_EXCEPTION(UnsupportedFeature()
            << MessageErrInfo("No support for negative shifts as it would involve discarding information"));
  }
  new_origin_ = IntPoint2D(xshift, yshift);

  t_pointset subtractPointSet;
  allPoints(orig_mmd.width, orig_mmd.height, new_origin_, subtractPointSet);

  BOOST_FOREACH( const IntPoint2D& point, subtractPointSet)
  {
    newPoints_.erase(point);
  }
}

void MapTransform::allPoints(unsigned int xsize, unsigned int ysize, t_pointset& points)
{
  for (unsigned int x = 0; x < xsize; ++x)
  {
    for (unsigned int y = 0; y < ysize; ++y)
    {
      points.insert(IntPoint2D(x, y));
    }
  }
}

void MapTransform::allPoints(unsigned int xsize, unsigned int ysize, const IntPoint2D& shift, t_pointset& points)
{
  for (unsigned int x = 0; x < xsize; ++x)
  {
    for (unsigned int y = 0; y < ysize; ++y)
    {
      points.insert(IntPoint2D(x + shift.x(), y + shift.y()));
    }
  }
}


bool MapTransform::needsResize() const { return needs_resize_; }
bool MapTransform::needsShift() const { return needs_shift_; }
const IntPoint2D& MapTransform::newOrigin() const { return new_origin_; }
const IntSize2D& MapTransform::shiftWindowSize() const { return shift_size_; }


std::auto_ptr<limits> limits::instance_;


