/*
 * opencv_tools.cpp
 *
 *  Created on: 11/08/2011
 *      Author: daver
 */

#include <ros/ros.h>
#include <sstream>
#include "opencv_tools.h"

namespace cvtools
{

CVManager::CVManager(): width_(0), height_(0), xorigin_(0), yorigin_(0)
{
  cv::startWindowThread();
}

CVManager::~CVManager()
{
  while (!name2mat_.empty())
  {
    t_name2mat::iterator it = name2mat_.begin();
    cv::destroyWindow(it->first);
    name2mat_.erase(it);
  }
}

std::auto_ptr<CVManager> CVManager::instance_;

CVManager& CVManager::instance()
{
  if (instance_.get() == NULL)
  {
    std::auto_ptr<CVManager> tmp (new CVManager());
    instance_ = tmp;
  }
  return *instance_;
}

void CVManager::resize(unsigned int width, unsigned int height, int xorigin, int yorigin)
{
  if (width_ == width && height_ == height &&
      xorigin_ == xorigin && yorigin_ == yorigin) return;
//      bool resize = (width_ != width || height_ != height);

  BOOST_FOREACH (t_name2mat::value_type& pair, name2mat_)
  {
    boost::shared_ptr<cv::Mat> img(new cv::Mat(height, width, CV_8UC3));
    cv::imshow(pair.first, *img); // Do this so I don't get memory corruption
    pair.second = img;
  }
  width_ = width;
  height_ = height;
  xorigin_ = xorigin;
  yorigin_ = yorigin;

  allpoints_.clear();
  for (unsigned int x = 0; x < width_; ++x)
  {
    for (unsigned int y = 0; y < height_; ++y)
    {
      allpoints_.push_back(IntPoint2D(x,y));
    }
  }
}

cv::Mat& CVManager::getImage(const std::string& name)
{
  t_name2mat::iterator it = name2mat_.find(name);
  if (it != name2mat_.end()) return *(it->second);

  ROS_DEBUG_STREAM("Creating new window: " << name);
  boost::shared_ptr<cv::Mat> mat(new cv::Mat(height_, width_, CV_8UC3));
  if (mat.get() == NULL || mat->data == NULL)
  {
    BOOST_THROW_EXCEPTION(UnsupportedFeature()
                          << MessageErrInfo("OpenCV error creating window"));
  }
//  cv::namedWindow(name, CV_WINDOW_AUTOSIZE);
//  cv::namedWindow(name, CV_WINDOW_NORMAL | CV_WINDOW_FREERATIO | CV_GUI_EXPANDED);
  cv::namedWindow(name, CV_WINDOW_NORMAL | CV_WINDOW_FREERATIO);

  name2mat_[name] = mat;

  unsigned long id = name2mat_.size();
  id2name_[id] = name;
  cv::setMouseCallback(name, CVManager::OnMouse, (void *) id);

  return *mat;
}

void CVManager::showImage(const std::string& name)
{
  t_name2mat::iterator it = name2mat_.find(name);
  if (it == name2mat_.end()) return;
  cv::imshow(name, *(it->second));
}

void CVManager::updateOccProbGrid(const t_occgrid& oc, const std::vector<IntPoint2D>& in_points)
{
  cvtools::OccProbGrid2RGB occ2rgb;
  cv::Mat& img = getImage("Occupancy Grid");
  copy(oc, in_points, img, occ2rgb);
  cv::imshow("Occupancy Grid", img);
}

void CVManager::updateOccProbGrid(const t_occgrid& oc)
{
  cvtools::OccProbGrid2RGB occ2rgb;
  cv::Mat& img = getImage("Occupancy Grid");
  copy(oc, img, occ2rgb);
  cv::imshow("Occupancy Grid", img);
}

void CVManager::updateOccMap(const t_occmap& oc, const std::vector<IntPoint2D>& in_points)
{
  cvtools::OccMap2RGB occ2rgb;
  cv::Mat& img = getImage("Occupancy Map");
  copy(oc, in_points, img, occ2rgb);
  cv::imshow("Occupancy Map", img);
}

void CVManager::updateOccMap(const t_occmap& oc)
{
  cvtools::OccMap2RGB occ2rgb;
  cv::Mat& img = getImage("Occupancy Map");
  copy(oc, img, occ2rgb);
  cv::imshow("Occupancy Map", img);
}

void CVManager::updateOccMap2(const t_occmap& oc, const std::vector<IntPoint2D>& in_points, const VoronoiRegionMap& rm)
{
  cvtools::OccMap2RGB occ2rgb;
  cv::Mat& img = getImage("Occupancy Map2");
//  copy(oc, in_points, img, occ2rgb);
  copy(oc, img, occ2rgb);

  regionmap_ = rm.clone();
  DisplayRegions("Occupancy Map2");

  cv::imshow("Occupancy Map2", img);
}

void CVManager::updateOccMap2(const t_occmap& oc, const VoronoiRegionMap& rm)
{
  cvtools::OccMap2RGB occ2rgb;
  cv::Mat& img = getImage("Occupancy Map2");
  copy(oc, img, occ2rgb);

  regionmap_ = rm.clone();
  DisplayRegions("Occupancy Map2");

  cv::imshow("Occupancy Map2", img);
}


void CVManager::updateBasicMap(const t_basicmap& oc, const std::vector<IntPoint2D>& in_points)
{
  cvtools::BasicMap2RGB filter;
  cv::Mat& img = getImage("Basic Map");
  copy(oc, in_points, img, filter);
  cv::imshow("Basic Map", img);
}

void CVManager::updateBasicMap(const t_basicmap& oc)
{
  cvtools::BasicMap2RGB filter;
  cv::Mat& img = getImage("Basic Map");
  copy(oc, img, filter);
  cv::imshow("Basic Map", img);
}

void CVManager::update(const t_voronoimap& oc, const VoronoiRegionMap& rm)
{
  vmap_ = create_grid2d<t_voronoimap, t_voronoimap>(oc);
  regionmap_ = rm.clone();

  cv::Mat& img = getImage("Voronoi Map");
  render_voronoimap(oc, allpoints_.begin(), allpoints_.end(), img);
  // FIXUP
  //DisplayRegions("Voronoi Map");
  cv::imshow("Voronoi Map", img);
}

void CVManager::OnMouse(int event, int x, int y, int flags, void *param)
{
  if (event != CV_EVENT_LBUTTONDOWN) return;

  unsigned int id = reinterpret_cast<unsigned long>(param);
  CVManager& cvm = CVManager::instance();

  t_id2name::iterator it = cvm.id2name_.find(id);
  if (it == cvm.id2name_.end())
  {
    BOOST_THROW_EXCEPTION(Exception()
                          << MessageErrInfo("OpenCV error creating window"));
  }
  const std::string& name = it->second;

  if (name.compare("Voronoi Map") != 0)
  {
    std::cout << "Ignoring mouse presses for anything other than \"Voronoi Map\"" << std::endl;
    return;
  }
  int vmap_height = cvm.vmap_->shape()[1];
  IntPoint2D point(x, (vmap_height-1) - y);
  cvm.DisplayCell(*(cvm.vmap_), point);
}

void CVManager::DisplayCell(const t_voronoimap& vmap, const IntPoint2D& point)
{
	//std::cerr << "VMAP: " << vmap.shape()[0] << " x " << vmap.shape()[1] << ", point: " << point << std::endl;
	//std::cerr << std::endl;
	if (point.x() < 0 || point.y() < 0 || point.x() >= vmap.shape()[0] || point.y() >= vmap.shape()[1]) return;

	const VoronoiCell &vc = vmap[point.x()][point.y()];

	if (vc.isVoronoi())
	{
		DisplayVoronoiCell(vc, point);
	}
	else
	{
		bool found = false;
		std::vector<IntPoint2D> adjlist;
		adjacents8(vmap, point, adjlist);

		BOOST_FOREACH(const IntPoint2D& ap, adjlist)
		{
			const VoronoiCell& vca = vmap[ap.x()][ap.y()];
			if (vca.isVoronoi())
			{
				DisplayVoronoiCell(vca, ap);
				found = true;
			}
		}
		if (!found)
		{
			std::cout << "Point " << point << ": ";
			if (vc.isFree()){ std::cout << "is free" << std::endl; }
			if (vc.isOccupied()){ std::cout << "is occupied" << std::endl; }
		}
	}
}

void drawRegion(cv::Mat& img, int image_height, const cv::Scalar& colour, const VoronoiRegion& region)
{
  if (!region.isValid()) return;

   cv::Point mp(region.midpoint_.x()-5, (image_height-1)-region.midpoint_.y());
   std::stringstream ss;
   int fontFace = cv::FONT_HERSHEY_COMPLEX;
   double fontScale = 0.45;
   float linewidth = 2.0;

   // FIXUP: FOR MORRI
   ss << region.id_;
   // cv::putText(img, ss.str(), mp, fontFace, fontScale, colour);

   // draw the region
   std::vector<IntPoint2D> poly = region.polygon_;
   if (poly.size() >= 2)
   {
     std::vector<IntPoint2D>::const_iterator it = poly.begin();
     std::vector<IntPoint2D>::const_iterator end = poly.end();
     cv::Point first(it->x(), (image_height-1) - it->y());
     while (it != end)
     {
       cv::Point p1(it->x(), (image_height-1) - it->y());
       if (++it == end)
       {
         if (p1 != first) cv::line(img, p1, first, colour,linewidth);
         break;
       }
       cv::Point p2(it->x(), (image_height-1) - it->y());
// FIXUP: FOR MORRI
       //       cv::line(img, p1, p2, colour);
       cv::line(img, p1, p2, colour,linewidth);
     }
   }
}

void CVManager::DisplayRegions(const std::string& imgname)
{
  cv::Scalar red = cv::Scalar(0,0,255); // Red?
  cv::Scalar green = cv::Scalar(0,255,0); // Green?
  cv::Scalar blue = cv::Scalar(255,0,0); // Blue?
  cv::Scalar yellow = cv::Scalar(0,255,255); // yellow
  cv::Scalar pink = cv::Scalar(255,0,255); // pink
  cv::Scalar cyan = cv::Scalar(255,255,0); //cyan
  cv::Scalar black = cv::Scalar(0,0,0); //black

  cv::Mat& img = getImage(imgname);
  //int vmap_height = vmap_->shape()[1];
  int map_height = img.rows;

  //if (map_height != vmap_height){ std::<<count "PROBLEMS" << std::endl;}

  // Display each region
  VoronoiRegionMap::t_p2r::const_iterator rcurr;
  VoronoiRegionMap::t_p2r::const_iterator rend;
  boost::tie(rcurr,rend) = regionmap_->regions();
  while (rcurr != rend)
  {
    const VoronoiRegion& region = *rcurr++->second;
    if (!region.isValid()) continue;
    drawRegion(img, map_height, blue, region);
  }
}

void CVManager::DisplayVoronoiCell(const VoronoiCell& vc, const IntPoint2D& point)
{
  int vmap_height = vmap_->shape()[1];

  if (!vc.isVoronoi()) return;
  if (vc.isCriticalCorridor()) std::cout << "Critical corridor point ";
  else if (vc.isCorridor()) std::cout << "Corridor point ";
  else std::cout << "Voronoi point ";

  int p2o1 = point.sqdistanceTo(vc.obst_);
  int p2o2 = point.sqdistanceTo(vc.obst2_);
  std::cout << point << " with obstacle " << vc.obst_ << " and " << vc.obst2_ << std::endl;
  std::cout << "\t sq distance to obstacles = " <<  p2o1 << " and " << p2o2 << std::endl;
  std::cout << "\t sq distance between obstacles = " << vc.obst_.sqdistanceTo(vc.obst2_)  << std::endl;
  std::cout << "\t distance between obstacles = " << vc.obst_.distanceTo(vc.obst2_)  << std::endl;
  std::cout << "\t Adjusted: obst1 = " << p2o1*3+ p2o2 << ", obst2 = " << p2o2*3+p2o1 << std::endl;
  if (vc.isCorridor())
  {
	  std::cout << "\t CorrVariance = " << vc.corrVariance_ << std::endl;
	  std::cout << "\t CCRatio = " << vc.ccRatio_ << std::endl;
  }

  cv::Mat& img = getImage("Voronoi Map");
  cv::Mat imgclone;
  img.copyTo(imgclone);

  // Draw the voronoi details
  cv::Scalar red = cv::Scalar(0,0,255); // Red?
  cv::Scalar green = cv::Scalar(0,255,0); // Green?
  cv::Scalar blue = cv::Scalar(255,0,0); // Blue?
  cv::Scalar yellow = cv::Scalar(0,255,255); // yellow
  cv::Scalar pink = cv::Scalar(255,0,255); // pink
  cv::Scalar cyan = cv::Scalar(255,255,0); //cyan


  cv::Point vp(point.x(), (vmap_height-1) - point.y());
  cv::Point obst1p(vc.obst_.x(), (vmap_height-1) - vc.obst_.y());
  cv::Point obst2p(vc.obst2_.x(), (vmap_height-1) - vc.obst2_.y());
  cv::line(imgclone, vp, obst1p, pink);
  cv::line(imgclone, vp, obst2p, pink);

  if (regionmap_->inRegion(point))
  {
    const VoronoiRegion& region = regionmap_->region(point);
    std::cout << "\t Region ID: " << region.id_ << std::endl;
    std::cout << "\t Midpoint: " << region.midpoint_ << std::endl;
    std::cout << "\t Related pointsets: ";

    //    cv::circle(imgclone,  mp, 1, red);
//    cv::circle(imgclone,  mp, 5, red);

    BOOST_FOREACH(const t_vpointmap::value_type& vp, region.vpoints_)
    {
        const IntPoint2D& psp = vp.first;
      cv::Point cvpsp(psp.x(), (vmap_height-1)-psp.y());
      //      cv::circle(imgclone,  cvpsp, 0, green);
      std::cout << psp << " ";
    }
    std::cout << std::endl;


    std::cout << "\t Intersecting regions: [";

    typedef std::set<std::pair<unsigned long, unsigned long> > t_tmp;
    t_tmp::const_iterator icurr = regionmap_->intersect_.begin();
    t_tmp::const_iterator iend = regionmap_->intersect_.end();

    if (regionmap_->intersect_.empty()) std::cout << "EMPTY";

    while (icurr != iend)
    {
        if (icurr->first == region.id_) std::cout << " " << icurr->second << " ";
        ++icurr;
    }
    std::cout << "]" << std::endl;

    drawRegion(imgclone, vmap_height, green, region);
  }

  std::cout << std::endl;
  cv::imshow("Voronoi Map", imgclone);
}

}
