#include <sstream>
#include <ros/ros.h>
#include <std_msgs/UInt32MultiArray.h>
#include <region_msgs/Region.h>
#include <region_msgs/Regions.h>
#include "regionmapper_serv.h"
#include "opencv_tools.h"
#include "voronoi_region_calc.h"

RegionMapperServ::RegionMapperServ()
{

  // Read all the relevant ROS paramaters
  ros::NodeHandle private_nh_("~");

  // The frame and subscribe/publish topics
  private_nh_.param("map_frame", map_frame_, std::string("map"));
  private_nh_.param("map", map_topic_, std::string("map"));
  private_nh_.param("regions_topic", regions_topic_, std::string("regions_topic"));
  private_nh_.param("regions_query", regions_query_, std::string("regions_query"));

  // parameters to control the region detection
  double min_obstacle_dist;
  double min_corridor_dist;
  double epsilon_neighbour_dist;
  double skeleton_obstacle_dist;

  private_nh_.param("min_obstacle_dist", min_obstacle_dist, double(0.6));
  private_nh_.param("min_corridor_dist", min_corridor_dist, double(0.6));
  private_nh_.param("epsilon_neigbour_dist", epsilon_neighbour_dist, double(1.0));
  private_nh_.param("skeleton_obstacle_dist", skeleton_obstacle_dist, double(1.0));
  private_nh_.param("gui", gui_, false);

  voronoimapupdater_.minObstDistance(min_obstacle_dist);
  voronoimapupdater_.minCorrDistance(min_corridor_dist);
  voronoimapupdater_.epNeighbourDistance(epsilon_neighbour_dist);
  voronoimapupdater_.skelObstDistance(skeleton_obstacle_dist);

  // Subscribe to the map topic
  ros::NodeHandle global_nh_;
  map_sub_ = global_nh_.subscribe(map_topic_, 1, &RegionMapperServ::updateOccGridCB, this);
  ROS_INFO_STREAM("Subscribing to occupancy grid topic " << map_topic_);

  // Advertise to publish the regions topic (this publishes the ids of regions)
  regionids_pub_ = global_nh_.advertise<std_msgs::UInt32MultiArray>(regions_topic_, 1, true);
  ROS_INFO_STREAM("Opening topic " << regions_topic_ << " for publishing region ids (std_msgs::UInt32MultiArray)");

  // Advertise the service to query regions
  regions_serv_ = global_nh_.advertiseService(regions_query_, &RegionMapperServ::regionsQuery,this);
}


RegionMapperServ::~RegionMapperServ()
{
  ROS_DEBUG("Destroying the RegionMapperServ");
}

/*
 * Check that we have a valid occupancy grid.
 * For the moment check that there is at least one occupied point.
 */

bool RegionMapperServ::_is_valid_occupancy_grid(const nav_msgs::OccupancyGridPtr &oc, int threshold)
{
	BOOST_FOREACH(char d, oc->data)
	{
		if (d >= threshold) return true;
	}
	ROS_ERROR("Nothing to do as occupancy grid has no occupied points");
	return false;
}


bool RegionMapperServ::regionsQuery(region_msgs::GetRegions::Request &req,
									region_msgs::GetRegions::Response &res)
{
	const geometry_msgs::Point& rorigin = mapmetadata_.origin.position;
	float resolution = mapmetadata_.resolution;
	std::set<unsigned int> idset;
	BOOST_FOREACH(unsigned int i, req.query){ idset.insert(i); }
	VoronoiRegionMap::t_p2r::const_iterator rcurr;
	VoronoiRegionMap::t_p2r::const_iterator rend;
	boost::tie(rcurr,rend) = voronoigrouper_.regionmap().regions();

	region_msgs::Regions rregions;
	int num_valid_regions = 0;
	while (rcurr != rend)
	{
		std::stringstream ss;
		const VoronoiRegion& vregion = *(rcurr++->second);
		if (!vregion.isValid()) continue;
		if (!idset.empty() && idset.count(vregion.id_) == 0) continue;
		++num_valid_regions;
		res.regions.push_back(region_msgs::Region());
		region_msgs::Region& rregion = res.regions.back();
		rregion.rtype = region_msgs::Region::FREE;
		rregion.polygon.holes.clear();
		rregion.labels.clear();

		rregion.header.seq = vregion.id_;
		rregion.header.frame_id = map_frame_id_;
		rregion.header.stamp = ros::Time::now();
		ROS_DEBUG_STREAM("Region: id=" << vregion.id_ << ", polygon size=" << vregion.polygon_.size());
		_rospolygon(vregion.polygon_, rorigin, resolution, rregion.polygon.outer.points);
	}
	ROS_INFO_STREAM("Returning " << num_valid_regions << " valid regions");
	return true;
}


void
RegionMapperServ::updateOccGridCB(const nav_msgs::OccupancyGridPtr &oc)
{
	try
	{
		if (!_is_valid_occupancy_grid(oc, 50)) return;
		if (occgrid_.get() == NULL)
			initMaps(oc);
		else
			updateMaps(oc);
	}
	catch (const boost::exception& e)
	{
		ROS_FATAL_STREAM("Exception: " << diagnostic_information(e));
		std::cerr << diagnostic_information(e) << std::endl;
		ros::shutdown();
	}
	catch (const std::exception& e)
	{
		ROS_FATAL_STREAM("Exception: " << e.what());
		std::cerr << e.what() << std::endl;
		ros::shutdown();
	}
	catch (...)
	{
		ROS_FATAL("Caught unknown exception. Shuting down");
		std::cerr << "Caught unknown exception. Shuting down." << std::endl;
		ros::shutdown();
	}
}

void
RegionMapperServ::initMaps(const nav_msgs::OccupancyGridPtr& oc)
{
  mapmetadata_ = oc->info;
  map_frame_id_ = oc->header.frame_id;
  ROS_INFO_STREAM("Initialising maps of width: " << mapmetadata_.width
		           << " and height: " << mapmetadata_.height);

  // Create a view of the occupancy grid message and fill the internal occupancy grid
  t_const_occgrid_ref oc_data(&(oc->data)[0],
      boost::extents[mapmetadata_.width][mapmetadata_.height],
      boost::fortran_storage_order());
  occgrid_ = create_grid2d<t_occgrid, t_const_occgrid_ref> (oc_data);

  // mapupdater_.resolution(mapmetadata_.resolution);
  // Create the map by transforming data from the occupancy grid and then guess unknowns where possible
  occmap_ = create_grid2d<t_occmap> (mapmetadata_.width, mapmetadata_.height);
  occmapupdater_.update(*occmap_, *occgrid_);

  // Create the voronoi map from the basic map
  //  voronoimap_ = create_grid2d<t_voronoimap, t_basicmap, BasicMap2VoronoiMap>(*basicmap_, voronoimapfilter_);
  t_pointset voronoimap_boundarypoints;

//  voronoimap_ = voronoimap_create(*basicmap_, voronoimap_boundarypoints);
  voronoimap_ = voronoimap_create(*occmap_, voronoimap_boundarypoints);
  voronoimapupdater_.resolution(mapmetadata_.resolution);
  voronoimapupdater_.update(*voronoimap_, voronoimap_boundarypoints.begin(), voronoimap_boundarypoints.end());
  //voronoimapupdater_.update(*voronoimap_);

  ROS_INFO("Computing regions from updated occupancy grid");
  voronoigrouper_.update(*voronoimap_, voronoimapupdater_.changes().begin(),
                                       voronoimapupdater_.changes().end());

  if (gui_)
  {
	  cvtools::CVManager& cvman = cvtools::CVManager::instance();
	  cvman.resize(mapmetadata_.width, mapmetadata_.height, 0, 0);
	  cvman.updateOccProbGrid(*occgrid_);

	  // TEMPORAY TO DISPLAY THE FILTERED OCCUPANCY MAP
	  cvman.updateOccMap(*occmap_);
	  cvman.updateOccMap2(*occmap_, voronoigrouper_.regionmap());
	  cvman.update(*voronoimap_, voronoigrouper_.regionmap());
  }

  // Publish the regions to ROS clients
  publishRegions(voronoigrouper_.regionmap(), mapmetadata_);
}

void RegionMapperServ::updateMaps(const nav_msgs::OccupancyGridPtr &oc)
{
  const nav_msgs::MapMetaData& oc_mmd = oc->info;

  // Create a grid2d view of the occupancy grid message and fill the internal occupancy grid
  t_const_occgrid_ref oc_data(&(oc->data)[0],
      boost::extents[oc_mmd.width][oc_mmd.height],
      boost::fortran_storage_order());

  std::vector<IntPoint2D> changes_occgrid;
  //std::vector<IntPoint2D> changes_basicmap;
  std::vector<IntPoint2D> changes_voronoimap;
  std::vector<IntPoint2D> changes_occmap;

  cvtools::CVManager & cvman = cvtools::CVManager::instance();
  MapTransform mt(mapmetadata_, oc_mmd);


  /* First resize all the grids if necessary */
  if (mt.needsResize())
  {
	  ROS_INFO_STREAM("Resizing maps to width: " << oc_mmd.width
		               << " and height: " << oc_mmd.height);

	std::auto_ptr<t_occgrid> n_occgrid = create_grid2d<t_occgrid> (oc_mmd.width, oc_mmd.height);
    std::auto_ptr<t_occmap> n_occmap = create_grid2d<t_occmap>(oc_mmd.width, oc_mmd.height);
    std::auto_ptr<t_voronoimap> n_voronoimap = create_grid2d<t_voronoimap>(oc_mmd.width, oc_mmd.height);
    grid2d_copy_to(*occgrid_, *n_occgrid, -1, mt.newOrigin());
    occmap_copy_to(*occmap_, *n_occmap, mt.newOrigin(), changes_occgrid);
    voronoimap_copy_to(*voronoimap_, *n_voronoimap, mt.newOrigin());
    occgrid_ = n_occgrid;
    occmap_ = n_occmap;
    voronoimap_ = n_voronoimap;
  }

  if (mt.needsResize() && gui_)
  {
	  cvman.resize(oc_mmd.width, oc_mmd.height, 0, 0);
	  cvman.updateOccProbGrid(*occgrid_);

	  // TEMPORAY TO DISPLAY THE FILTERED OCCUPANCY MAP
	  cvman.updateOccMap(*occmap_);
	  cvman.updateOccMap2(*occmap_, voronoigrouper_.regionmap());

	  cvman.update(*voronoimap_, voronoigrouper_.regionmap());
  }

  mapmetadata_ = oc_mmd;

  // Update the occupancy grid and return the list of points that have changed
  grid2d_update(oc_data, *occgrid_, changes_occgrid, PassThroughFunctor<signed char> ());

  // Update the occupancy map and save list of changed points
  occmapupdater_.update(*occmap_, *occgrid_, changes_occgrid.begin(), changes_occgrid.end());
  BOOST_FOREACH(const IntPoint2D& p, occmapupdater_.changes()) changes_occmap.push_back(p);

//  grid2d_update(*basicmap_, changes_basicmap, *voronoimap_, changes_voronoimap, voronoimapfilter_);
  grid2d_update(*occmap_, changes_occmap, *voronoimap_, changes_voronoimap, voronoimapfilter_);

  // Update the voronoi map and print out the percentage of changes made
  voronoimapupdater_.update(*voronoimap_, changes_voronoimap.begin(), changes_voronoimap.end());

  ROS_INFO("Computing regions from updated occupancy grid");
  ROS_DEBUG_STREAM("Occupancy grid update: " << changes_occgrid.size() << " points ("
		           << (changes_occgrid.size() * 100) / occgrid_->num_elements() << "% of total)");
  ROS_DEBUG_STREAM("Occupancy map update: " << changes_occmap.size() << " points (" <<
		           (changes_occmap.size() * 100) / occmap_->num_elements() << "% of total)");
  ROS_DEBUG_STREAM("Voronoi map update: " << voronoimapupdater_.changes().size() << " points (" <<
		           (voronoimapupdater_.changes().size() * 100) / voronoimap_->num_elements() << "% of total)");

  if (mt.needsResize())
  {
    // FIXUP: Ideally I should shift all points in the regionMap as a
    // result of the resize. However, for the moment recompute all the regions.
    voronoigrouper_.update(*voronoimap_);

    if (gui_)
    {
    	cvman.resize(oc_mmd.width, oc_mmd.height, 0, 0);
    	cvman.updateOccProbGrid(*occgrid_);

    	// TEMPORAY TO DISPLAY THE FILTERED OCCUPANCY MAP
    	cvman.updateOccMap(*occmap_);
    	cvman.updateOccMap2(*occmap_, voronoigrouper_.regionmap());

    	cvman.update(*voronoimap_, voronoigrouper_.regionmap());
    }
  }
  else
  {
    voronoigrouper_.update(*voronoimap_, voronoimapupdater_.changes().begin(),
                                         voronoimapupdater_.changes().end());

    if (gui_)
    {
    	cvman.updateOccProbGrid(*occgrid_, changes_occgrid);

    	// TEMPORAY TO DISPLAY THE FILTERED OCCUPANCY MAP
    	cvman.updateOccMap(*occmap_, changes_occmap);
    	cvman.updateOccMap2(*occmap_, changes_occmap, voronoigrouper_.regionmap());
    	cvman.update(*voronoimap_, voronoigrouper_.regionmap());
    }
  }

  // Publish the regions to the ROS region topic
  publishRegions(voronoigrouper_.regionmap(), mapmetadata_);
}

region_msgs::Point2D RegionMapperServ::_rospoint2rospoint32(const geometry_msgs::Point& point)
{
  region_msgs::Point2D p32;
  p32.x = static_cast<float>(point.x);
  p32.y = static_cast<float>(point.y);
  return p32;
}

geometry_msgs::Point RegionMapperServ::_rospoint(const IntPoint2D& inpoint,
												 const geometry_msgs::Point& origin,
												 float resolution)
{
  geometry_msgs::Point point;
  point.x = origin.x + inpoint.x()*resolution;
  point.y = origin.y + inpoint.y()*resolution;
  point.z = origin.z;
  return point;
}

void RegionMapperServ::_rospolygon(const std::vector<IntPoint2D>& points,
								   const geometry_msgs::Point& origin, float resolution,
								   std::vector<region_msgs::Point2D>& rospoints)
{
  std::vector<IntPoint2D>::const_iterator curr = points.begin();
  std::vector<IntPoint2D>::const_iterator end = points.end();
  while (curr != end)
  {
    geometry_msgs::Point point64 = _rospoint(*curr++, origin, resolution);
    rospoints.push_back(_rospoint2rospoint32(point64));
  }
}

void RegionMapperServ::publishRegions(const VoronoiRegionMap& vrm, const nav_msgs::MapMetaData& mmd)
{
	VoronoiRegionMap::t_p2r::const_iterator rcurr;
	VoronoiRegionMap::t_p2r::const_iterator rend;
	boost::tie(rcurr,rend) = vrm.regions();

	std_msgs::UInt32MultiArray regionids;
	int num_valid_regions = 0;
	while (rcurr != rend)
	{
		std::stringstream ss;
		const VoronoiRegion& vregion = *(rcurr++->second);
		if (!vregion.isValid()) continue;
		++num_valid_regions;
		regionids.data.push_back(vregion.id_);
	}
	ROS_INFO_STREAM("Identified " << num_valid_regions << " valid regions");
	regionids_pub_.publish(regionids);

	/*

  const geometry_msgs::Point& rorigin = mmd.origin.position;
  float resolution = mmd.resolution;
  VoronoiRegionMap::t_p2r::const_iterator rcurr;
  VoronoiRegionMap::t_p2r::const_iterator rend;
  boost::tie(rcurr,rend) = vrm.regions();

  region_msgs::Regions rregions;
  int num_valid_regions = 0;
  while (rcurr != rend)
  {
    std::stringstream ss;
    const VoronoiRegion& vregion = *(rcurr++->second);
    if (!vregion.isValid()) continue;
    ++num_valid_regions;
    rregions.regions.push_back(region_msgs::Region());
    region_msgs::Region& rregion = rregions.regions.back();
    rregion.rtype = region_msgs::Region::FREE;
    rregion.polygon.holes.clear();
    rregion.labels.clear();

    rregion.header.seq = vregion.id_;
    rregion.header.frame_id = map_frame_id_;
	rregion.header.stamp = ros::Time::now();
    ROS_DEBUG_STREAM("Region: id=" << vregion.id_ << ", polygon size=" << vregion.polygon_.size());
    _rospolygon(vregion.polygon_, rorigin, resolution, rregion.polygon.outer.points);
  }
  ROS_INFO_STREAM("Identified " << num_valid_regions << " valid regions");
  regionids_pub_.publish(rregions);
  */
}



