/*
 * voronoi_region_calc.cpp
 *
 *  Created on: 17/10/2011
 *      Author: daver
 */

#include <cmath>
#include <CGAL/Cartesian.h>
#include <CGAL/Polygon_2.h>
#include <CGAL/Polygon_2_algorithms.h>
#include <CGAL/Polygon_with_holes_2.h>
#include <CGAL/Polygon_set_2.h>
#include <CGAL/convex_hull_2.h>
#include <CGAL/Boolean_set_operations_2.h>
#include <CGAL/Alpha_shape_2.h>
#include <ros/ros.h>

#include "voronoi_region_calc.h"
#include "grid2d.h"

typedef CGAL::Cartesian<int> CK;
typedef CK::Point_2 CPoint;
typedef CGAL::Polygon_2<CK> CPolygon;
typedef CGAL::Polygon_with_holes_2<CK> CPolygonWithHoles;
typedef CGAL::Polygon_set_2<CK> CPolygonSet;
typedef boost::shared_ptr<CPolygon> t_polygon_sptr;

/*
  typedef CGAL::Alpha_shape_vertex_base_2<K> Av;
  typedef CGAL::Triangulation_face_base_2<K> Tf;
  typedef CGAL::Triangulation_default_data_structure_2<K,Av,Af> Tds;
  typedef CGAL::Delaunay_triangulation_2<K,Tds> Dt;
  typedef CGAL::Alpha_shape_2<Dt> Alpha_shape_2;
*/

/*****************************************************************************/

CPoint pconvert(const IntPoint2D& p)
{
    return CPoint(p.x(), p.y());
}

IntPoint2D pconvert(const CPoint& p)
{
    return IntPoint2D(p.x(), p.y());
}

/*****************************************************************************/

VoronoiRegion::VoronoiRegion(const IntPoint2D& p, double d) :
    midpoint_(p), dist_(0), id_(0)
{
    vpoints_.insert(std::make_pair(p, d));
}
VoronoiRegion::VoronoiRegion(const VoronoiRegion& r):
    midpoint_(r.midpoint_), dist_(r.dist_), vpoints_(r.vpoints_),
    ccpoints_(r.ccpoints_), polygon_(r.polygon_), id_(r.id_)
{ }

bool VoronoiRegion::isValid() const
{
    return (id_ > 0 && polygon_.size() > 0);
}

unsigned long VoronoiRegion::sm_nextid_ = 1;

/*****************************************************************************/

class RegionFiller
{
public:
    typedef BushFire2<int, IntPoint2D> t_bushfire;

private:
    struct Status
    {
        int maxdist_;
        bool crossedCC_; // has a critical corridor line been crossed?
        Status(int dist, bool crossed): maxdist_(dist), crossedCC_(crossed){}
        Status(): maxdist_(0), crossedCC_(false){}
    }; // The maximum distance that the bushfire can spread from this point
    typedef std::tr1::unordered_map<IntPoint2D, Status> t_p2c;
    t_p2c p2c_;
    t_pointset ccbpoints_; // critical corridor boundary points

    t_bushfire bushfire_;
    const t_voronoimap& grid_;
    std::vector<IntPoint2D> adjacents_;

    void findBoundaryPoints(t_pointset& boundarypoints);
    bool findPolygon(const t_pointset& boundarypoints,
                     const IntPoint2D& inregion,
                     std::vector<IntPoint2D>& polygon);
    bool buildConsecutivePoints(const IntPoint2D& start,
                                const t_pointset& boundarypoints,
                                std::vector<IntPoint2D>& vect);

public:
    RegionFiller(const t_voronoimap& grid);
    void burn(t_bushfire& bf, const int& distance, const IntPoint2D& point);
    void calculate(VoronoiRegion& region);
};


RegionFiller::RegionFiller(const t_voronoimap& grid): grid_(grid)
{ }

void RegionFiller::findBoundaryPoints(t_pointset& boundarypoints)
{
    // Find the boundary points
    BOOST_FOREACH(t_p2c::value_type& pair, p2c_)
    {
        adjacents_.clear();
        adjacents4(grid_, pair.first, adjacents_);
        BOOST_FOREACH(const IntPoint2D& p, adjacents_)
        {
            if (p2c_.find(p) == p2c_.end())
            {
                boundarypoints.insert(pair.first);
                break;
            }
        }
    }
}

// build consecutive points until a loop is closed
bool RegionFiller::buildConsecutivePoints(const IntPoint2D& start,
                                          const t_pointset& boundarypoints,
                                          std::vector<IntPoint2D>& vect)
{
    t_pointset visited;
    vect.push_back(start);
    bool done = false;
    int pathsize = 0;
    IntPoint2D next = start;
    while (!done)
    {
        std::vector<IntPoint2D> adjacents;
        adjacents8(next, adjacents);
        bool found = false;
        BOOST_FOREACH(next, adjacents)
        {
            if (visited.find(next) != visited.end()) continue;
            if (boundarypoints.find(next) == boundarypoints.end()) continue;
            if (next == start && pathsize <= 1) continue;
            if (next == start) done = true;
            visited.insert(next);
            ++pathsize;
            found = true;
            break;
        }
        if (found)
        {
            vect.push_back(next);
        }
        else if (!vect.empty())
        {
            next = *(vect.rbegin());
            vect.pop_back();
            --pathsize;
        }
        else
        {
            return false;
        }
    }
    return true;
}

/*

  bool build_alpha_shape(const t_pointset& boundarypoints, std::vector<IntPoint2D>& polygon)
  {
  std::vector<CPoint> cpoints;
  // Check if the inregion point is in the polygon - if yes then were done
  BOOST_FOREACH(IntPoint2D p, boundarypoints) cpoints.push_back(pconvert(p));

  Alpha_shape_2 A(cpoints.begin(), cpoints.end());
  A.set_mode(Alpha_shape_2::REGULARIZED);
  Alpha_iterator opt = A.find_optimal_alpha(1);
  A.set_alpha(*opt);
  Alpha_shape_2:: A.Alpha_shape_vertices_begin(); A.Alpha_shape_vertices_begin();

  for(Alpha_shape_edges_iterator it = A.alpha_shape_edges_begin();
  it != A.alpha_shape_edges_end();
  ++it){
  *out++ = A.segment(*it);
  }

  }
*/

bool RegionFiller::findPolygon(const t_pointset& boundarypoints,
                               const IntPoint2D& inregion,
                               std::vector<IntPoint2D>& polygon)
{
    CPoint pointinregion(pconvert(inregion));

    // Try out different starting points to see if we
    // can find a closed loop
    BOOST_FOREACH(const IntPoint2D& p, boundarypoints)
    {
        polygon.clear();
        if (buildConsecutivePoints(p, boundarypoints, polygon))
        {
            std::vector<CPoint> cpoints;
            // Check if the inregion point is in the polygon - if yes then were done
            BOOST_FOREACH(IntPoint2D p, polygon) cpoints.push_back(pconvert(p));
            /*
              if ((CGAL::bounded_side_2(cpoints.begin(), cpoints.end(),
              pointinregion, CK()) == CGAL::ON_BOUNDED_SIDE) &&
              (CGAL::is_simple_2(cpoints.begin(), cpoints.end()) ||
              CGAL::is_simple_2(cpoints.rbegin(), cpoints.rend()))) return true;
            */
            if ((CGAL::bounded_side_2(cpoints.begin(), cpoints.end(),
                                      pointinregion, CK()) == CGAL::ON_BOUNDED_SIDE)) return true;

        }
    }
    return false;
}

void RegionFiller::calculate(VoronoiRegion& region)
{
    p2c_.clear();
    ccbpoints_.clear();
    adjacents_.clear();
    bushfire_.clear();

    // Don't bother with single voronoi point regions as these are either
    // uninteresting or will be expanded as the robot explores further
    if (region.vpoints_.size() <= 1) return;


    // Keep track of the points along the line between
    // the criticial corridor points
    BOOST_FOREACH(const IntPoint2D& p, region.ccpoints_)
    {
        const VoronoiCell& cell = get_cell(grid_, p);
        std::list<IntPoint2D> line;
        p.pointsTo(cell.obst_, line, true);
        p.pointsTo(cell.obst2_, line, true);
        BOOST_FOREACH(const IntPoint2D& p2, line)
        {
            ccbpoints_.insert(p2);
        }
    }

    // Add the critical corridor and voronoi points
    // to start the bushfire. Note for the voronoi points
    // we have to be careful to check if it is on the
    // critical boundary line and if so adjust the
    // rate of expansion to be the same as the
    // critical boundary points.
    BOOST_FOREACH(const IntPoint2D& p, region.ccpoints_)
    {
        const VoronoiCell& cell = get_cell(grid_, p);
        int maxdist = std::ceil(std::sqrt(cell.sqdist_));
        //maxdist = maxdist *1.5;
        p2c_.insert(std::make_pair(p, Status(maxdist, true)));
        bushfire_.push(maxdist,p);
    }
    BOOST_FOREACH(const t_vpointmap::value_type& vp, region.vpoints_)
    {
        const IntPoint2D& p = vp.first;
        const VoronoiCell& cell = get_cell(grid_, vp.first);

        // FIXUP DPR20130326 - need to do a better job of calculating the expansion distance
        int maxdist = std::ceil(std::max(p.distanceTo(cell.obst_), p.distanceTo(cell.obst2_))*1.1);
//    int maxdist = std::ceil(std::sqrt(cell.sqdist_));
        if (ccbpoints_.find(p) == ccbpoints_.end())
            p2c_.insert(std::make_pair(p, Status(maxdist, false)));
        else
            p2c_.insert(std::make_pair(p, Status(maxdist, true)));
        bushfire_.push(maxdist,p);
    }

    bushfire_.update(t_bushfire::make_callback(this, &RegionFiller::burn));
    t_pointset boundarypoints;

    // Find the boundary points
    findBoundaryPoints(boundarypoints);

    // FIXUP JUST TO SEE THE DIFFERENCE
    //BOOST_FOREACH(const IntPoint2D& p, boundarypoints) region.polygon_.push_back(p);
    //return;

    // Determine the polygons within the boundary. The polygon detection may fail
    // for small regions (containing a single voronoi point). These are not interesting
    // anyway so ignore them.
    // FIXUP: The region may contain holes as a result of small internal obstacles.
    // At the moment I only look for the main region (by finding the simple polygon
    // that contains the region's midpoint). However, ultimately,
    // a hole should be detected and converted to some sort of occupied region.
    if (findPolygon(boundarypoints, region.midpoint_, region.polygon_))
    {
        region.id_ = VoronoiRegion::sm_nextid_++;
    }
    else
    {
        region.id_ = 0;
        region.polygon_.clear();
    }
}

void RegionFiller::burn(t_bushfire& bf, const int& distance, const IntPoint2D& point)
{
    Status& st = p2c_[point];
    if (distance != st.maxdist_) return;
    if (distance <= 1) return;
    int newdist = distance-1;
    bool ccboundary = st.crossedCC_;

    adjacents_.clear();
    adjacents4(grid_, point, adjacents_);
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
        const VoronoiCell& ac = get_cell(grid_, ap);
        if (!ac.isFree()) continue;
        if (!ccboundary) ccboundary = (ccbpoints_.find(ap) != ccbpoints_.end());
        if (ccboundary) newdist = std::ceil(static_cast<float>(newdist)*0.75);
        t_p2c::iterator it = p2c_.find(ap);
        if (it == p2c_.end())
        {
            p2c_[ap] = Status(newdist, ccboundary);
            bushfire_.push(newdist, ap);
        }
        else if (newdist > it->second.maxdist_)
        {
            it->second.maxdist_ = newdist;
            it->second.crossedCC_ = ccboundary;
            bushfire_.push(newdist, ap);
        }
    }
}


/*****************************************************************************/

std::pair<VoronoiRegionMap::t_p2r::const_iterator, VoronoiRegionMap::t_p2r::const_iterator>
VoronoiRegionMap::regions() const
{
    return make_pair(mp2r_.begin(), mp2r_.end());
}

size_t VoronoiRegionMap::numRegions() const
{
	return mp2r_.size();
}


void VoronoiRegionMap::newRegion(const IntPoint2D& vpoint, double distance)
{
    // point mustn't be part of a region
    BOOST_ASSERT(p2r_.find(vpoint) == p2r_.end());
    t_region_sptr region(new VoronoiRegion(vpoint, distance));
    p2r_[vpoint] = region;
}

void VoronoiRegionMap::newRegion(const IntPoint2D& vpoint, double distance,
                                 const IntPoint2D& ccpoint)
{
    // point mustn't be part of a region
    BOOST_ASSERT(p2r_.find(vpoint) == p2r_.end());
    t_region_sptr region(new VoronoiRegion(vpoint, distance));
    region->ccpoints_.insert(ccpoint);
    cc2r_[ccpoint] = region;
    p2r_[vpoint] = region;
}



void VoronoiRegionMap::removePoint(const IntPoint2D& point)
{
	// if its not part of a region do nothing
	t_p2r::iterator it = p2r_.find(point);
	if (it == p2r_.end()) return;
	t_region_sptr regptr = it->second;
	p2r_.erase(point);
	regptr->vpoints_.erase(point);
}

bool VoronoiRegionMap::removeRegion(const IntPoint2D& point)
{
	// if its not part of a region do nothing
	t_p2r::iterator it = p2r_.find(point);
	if (it == p2r_.end())
		if ((it = cc2r_.find(point)) == cc2r_.end()) return false;
	t_region_sptr regptr = it->second;
	ROS_DEBUG_STREAM("Region " << regptr->id_ << " removed due to change at cell " << point);
	mp2r_.erase(regptr->midpoint_);
	BOOST_FOREACH(const t_vpointmap::value_type& vp, regptr->vpoints_)
	{
        const IntPoint2D& p = vp.first;
		p2r_.erase(p);
	}
	BOOST_FOREACH(const IntPoint2D& ccp, regptr->ccpoints_)
	{
		cc2r_.erase(ccp);
	}
	if (regptr->id_ != 0) return true;
	return false;
}

void VoronoiRegionMap::addPoint(const IntPoint2D& regionpoint, 
                                const IntPoint2D& newpoint, double vdist,
                                const IntPoint2D& newendpoint, unsigned int distance)
{
    t_region_sptr& rsptr = p2r_[regionpoint];
    BOOST_ASSERT(rsptr->vpoints_.find(newendpoint) != rsptr->vpoints_.end());
    BOOST_ASSERT(distance > 0);
    rsptr->vpoints_.insert(std::make_pair(newpoint, vdist));
    p2r_[newpoint] = rsptr;
    if (distance > rsptr->dist_)
    {
        mp2r_.erase(rsptr->midpoint_);
        mp2r_[newpoint] = rsptr;
        rsptr->midpoint_ = newpoint;
        rsptr->dist_ = distance;
    }
}


void VoronoiRegionMap::mergeRegions(const IntPoint2D& a, const IntPoint2D& b)
{
    // Get the regions for the points
    t_p2r::iterator ita = p2r_.find(a);
    t_p2r::iterator itb = p2r_.find(b);

    BOOST_ASSERT(ita != p2r_.end() && itb != p2r_.end());

    // If the regions are the same do nothing otherwise
    // arbitrarily pick a's set as the one to keep.
    t_region_sptr psa = ita->second;
    t_region_sptr psb = itb->second;
    if (psa == psb) return;

    // Pick the midpoint that has the greatest distance
    if (psb->dist_ > psa->dist_)
    {
        mp2r_.erase(psa->midpoint_);
        mp2r_[psb->midpoint_] = psa;
        psa->dist_ = psb->dist_;
        psa->midpoint_ = psb->midpoint_;
    }
    else
    {
        mp2r_.erase(psb->midpoint_);
    }
    BOOST_FOREACH(const t_vpointmap::value_type& vp, psb->vpoints_)
    {
        const IntPoint2D& p = vp.first;
        psa->vpoints_.insert(vp);
        p2r_[p] = psa;
    }
    BOOST_FOREACH(const IntPoint2D& p, psb->ccpoints_)
    {
        psa->ccpoints_.insert(p);
        cc2r_[p] = psa;

    }
}

bool VoronoiRegionMap::inRegion(const IntPoint2D& point)
{
    return (p2r_.find(point) != p2r_.end());
}

bool VoronoiRegionMap::inRegionBoundary(const IntPoint2D& point)
{
    return (cc2r_.find(point) != cc2r_.end());
}


bool VoronoiRegionMap::sameRegion(const IntPoint2D& a, const IntPoint2D& b)
{
    // Get the regions for the points (if any)
    t_p2r::iterator ita = p2r_.find(a);
    t_p2r::iterator itb = p2r_.find(b);

    if (ita == p2r_.end()) return false;
    if (itb == p2r_.end()) return false;

    t_region_sptr psa = ita->second;
    t_region_sptr psb = itb->second;

    return (psa == psb);
}

std::auto_ptr<VoronoiRegionMap> VoronoiRegionMap::clone() const
{
    std::auto_ptr<VoronoiRegionMap> nrm(new VoronoiRegionMap());

    t_p2r::const_iterator oit;
    for (oit = p2r_.begin(); oit != p2r_.end(); oit++)
    {
        const IntPoint2D point = oit->first;
        t_region_sptr region = oit->second;
        t_p2r::iterator nit = nrm->p2r_.find(point);
        if (nit != nrm->p2r_.end()) continue;
        t_region_sptr nr(new VoronoiRegion(*region));

        BOOST_FOREACH(const t_vpointmap::value_type& vp, region->vpoints_)
        {
            const IntPoint2D& np = vp.first;
            nrm->p2r_[np] = nr;
        }
        nrm->mp2r_[nr->midpoint_] = nr; // copy to the midpoint map

        /*
          BOOST_FOREACH(const IntPoint2D& ccp, region->ccpoints_)
          {
          nr->ccpoints_.insert(ccp);
          }
          BOOST_FOREACH(const IntPoint2D& bp, region->polygon_)
          {
          nr->polygon_.push_back(bp);
          }
        */
    }
    nrm->intersect_ = intersect_;
    return nrm;
}


/*****************************************************************************/

VoronoiRegionCalc::VoronoiRegionCalc(): propagate_(true)
{ }

t_voronoimap& VoronoiRegionCalc::grid(){ return *gridptr_; }

// Relies on the earlier pruning of the diagram
bool VoronoiRegionCalc::isEndPoint(const IntPoint2D& p)
{
    adjacents_.clear();
    adjacents4(grid(), p, adjacents_);
    unsigned int numadj = 0;
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
        VoronoiCell&ac = get_cell(grid(),ap);
        if (ac.isVoronoi()) numadj++;
    }
    return (numadj <= 1);
}

// Only end voronoi points and critical corridors at the start of propagation
void VoronoiRegionCalc::setupPropagator()
{
	size_t num_removed = 0;
	size_t num_init_regions = regionmap_.numRegions();

	// Two passes - first remove any regions that need to be changed
	BOOST_FOREACH(const IntPoint2D& p, changedpoints_)
	{
		adjacents_.clear();
		adjacents4(grid(), p, adjacents_);
		adjacents_.push_back(p);
		BOOST_FOREACH(const IntPoint2D& p2, adjacents_)
		{
			VoronoiCell&c = get_cell(grid(), p2);
			bool remove = false;
			if (!regionmap_.inRegion(p2) && !regionmap_.inRegionBoundary(p2)) continue;
			if (!c.isVoronoi()) remove = true;
			if (c.isCriticalCorridor()) remove = true;
			if (regionmap_.inRegionBoundary(p2)) remove = true;
			if (remove)
			{
				if (regionmap_.removeRegion(p2)) num_removed ++; // Remove any corresponding regions
			}
			else
			{
				regionmap_.removePoint(p2);
				changedpoints_.erase(p2);
			}
		}
	}
	if (num_removed > 0)
		ROS_INFO_STREAM("Regions changed: removed "<< num_removed << " regions from " << num_init_regions);

    // Now setup the process to rebuild the new regions
    t_pointset visited;
    BOOST_FOREACH(const IntPoint2D& p, changedpoints_)
    {
        if (visited.find(p) != visited.end()) continue;
        VoronoiCell&c = get_cell(grid(), p);
        if (!c.isVoronoi()) continue;
        if (c.isCriticalCorridor())
        {
            adjacents_.clear();
            adjacents4(grid(), p, adjacents_);
            BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
            {
                if (visited.find(ap) != visited.end()) continue;
                const VoronoiCell& ac = get_cell(grid(),ap);
                if (!ac.isVoronoi()) continue;
                if (ac.isCriticalCorridor()) continue;
                if (regionmap_.inRegion(ap)) continue;

                // I think there is a special case where the corridor point is an
                // endpoint or is next to an endpoint so might be part of a region.
                // BOOST_ASSERT(isEndPoint(p) || isEndPoint(ap) || !regionmap_.inRegion(ap));
                p2nep_[ap] = EndPoint(ap,0);
                propagate_.push(0, ap);
                regionmap_.newRegion(ap, sqrt(ac.sqdist_), p);
                visited.insert(ap);
            }
        }
        else if (isEndPoint(p) && !regionmap_.inRegion(p))
        {
            p2nep_[p] = EndPoint(p,0);
            propagate_.push(0, p);
            regionmap_.newRegion(p, sqrt(c.sqdist_));
            visited.insert(p);
        }
    }
}

const VoronoiRegionMap& VoronoiRegionCalc::regionmap() const
{
    return regionmap_;
}

void VoronoiRegionCalc::process(t_bushfire2& bf, const int& distance, const IntPoint2D& point)
{
    VoronoiCell& cell = get_cell(grid(), point);
    BOOST_ASSERT(distance == 0 || !cell.isCriticalCorridor());
    EndPoint& nendpoint = p2nep_[point];

    adjacents_.clear();
    adjacents4(grid(), point, adjacents_);
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
        const VoronoiCell& ac = get_cell(grid(),ap);
        if (!ac.isVoronoi()) continue;
        if (ac.isCriticalCorridor()) continue;
        if (regionmap_.inRegion(ap))
        {
            regionmap_.mergeRegions(point, ap);
        }
        else
        {
            regionmap_.addPoint(point, ap, sqrt(ac.sqdist_), nendpoint.p_, distance+1);
            p2nep_[ap] = EndPoint(nendpoint.p_, distance+1);
            propagate_.push(distance+1, ap);
        }
    }
}

void VoronoiRegionCalc::buildBoundingPolygons()
{
	VoronoiRegionMap::t_p2r::const_iterator rcurr, rend;
	boost::tie(rcurr, rend) = regionmap_.regions();
	while (rcurr != rend)
	{
		VoronoiRegion& vregion = *((rcurr++)->second);
		if (vregion.isValid())
		{
			continue; // Already a valid (completed) region so do nothing.
		}
		RegionFiller rf(*gridptr_);
		rf.calculate(vregion);
	}
}

void VoronoiRegionCalc::calculate(t_voronoimap& grid)
{
    gridptr_ = &grid;
    setupPropagator();

    // First separate the regions in terms of groups of voronoi points.
    propagate_.update(t_bushfire2::make_callback(this, &VoronoiRegionCalc::process));

    // Second from these grouped voronoi points generate the bounding polygon
    buildBoundingPolygons();
}

void VoronoiRegionCalc::update(t_voronoimap& grid)
{
    std::vector<IntPoint2D> points;
    all_points(grid, points);
    update(grid, points.begin(), points.end());
}

void VoronoiRegionCalc::calculate_intersecting_regions()
{
    // To find the intersecting regions we iterate through all the critical
    // corridor points of each region look at the voronoi points on either
    // side to see which other region own them.
    regionmap_.intersect_.clear();
    VoronoiRegionMap::t_p2r::const_iterator rcurr;
    VoronoiRegionMap::t_p2r::const_iterator rend;
    boost::tie(rcurr,rend) = regionmap_.regions();

    while (rcurr != rend)
    {
        const VoronoiRegion& vregion = *(rcurr++->second);
        BOOST_FOREACH(const IntPoint2D& p, vregion.ccpoints_)
        {
            adjacents_.clear();
            adjacents24(grid(), p, adjacents_);
            BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
            {
                const VoronoiCell& ac = get_cell(grid(),ap);
                if (!ac.isVoronoi()) continue;
                VoronoiRegionMap::t_p2r::const_iterator it = regionmap_.p2r_.find(ap);
//                BOOST_ASSERT(it != regionmap_.p2r_.end());
                if (it == regionmap_.p2r_.end()) continue;
                if (it->second->id_ == vregion.id_) continue;
                regionmap_.intersect_.insert(std::make_pair(it->second->id_,
                                                            vregion.id_));
                regionmap_.intersect_.insert(std::make_pair(vregion.id_,
                                                            it->second->id_));
            }
        }
//        ++rcurr;
    }
}


